//**********************************************************
// Program: damages.java
//
// Author: Charles Marquet, cmm363@cornell.edu
//
// Description: This program calculates the environmental damages for all generation units.
//              
//
// Input: -Constants (from Abt and latimer Associates)
//
//        -EIA-EPA matched data 
//             Columns A-BK is EIA data
//             Columns BL-CP is stack data
//             Column CQ is EIA fuel category per unit
//             Columns CR-EL is EPA data matched by match.java
//             Column EM is the fips code for each EIA units
//
//        -Population data 
//             US census data by county matched with abt index for counties 1-3080 using file SR_matrix_readme)
//             Errors found in file: county 12025 is 12086, 29193 is 29186
//
//        -Transfer coeficients (from Abt, modified to remove errors)
//
//        -State % of population over 30 (from US census data)
//
// Need the free csv reader/writer from: http://sourceforge.net/projects/javacsv/
//*********************************************************

import java.io.*;
import java.lang.String;
import java.lang.Math;
import java.lang.Integer;
import com.csvreader.CsvReader;
import com.csvreader.CsvWriter;
import java.util.regex.*;
import java.util.*;


public class damages {
  //these are the input file names
  //Input: EIA-EPA-stacks matched file
  private static final String ADDRESS_FILE="EIA-EV-emissions_Gens.csv";//"match_NE_fips.csv";  
  //Input: Transfer Coeficients 
  private static final String ADDRESS_FILE2="TC.csv";
  //Input: population data by county
  private static final String ADDRESS_FILE3="pop.csv";
  //Input: % of population 30+ years by state
  private static final String ADDRESS_FILE4="state.csv";
  
  public static void main(String[] args) throws IOException {
    
    //these read the input files
    CsvReader reader = new CsvReader(new FileReader(ADDRESS_FILE));
    CsvReader reader2= new CsvReader(new FileReader(ADDRESS_FILE2));
    CsvReader reader3= new CsvReader(new FileReader(ADDRESS_FILE3));
    CsvReader reader4= new CsvReader(new FileReader(ADDRESS_FILE4));
    
    //constants
    Double beta =0.0046257;
    Double y =0.0082;
    Double noxcoef =1.7415;
    Double so2coef =2.065;
    Double life =7400000.0;
    Double secperyear=31536000.0;
    Double mgperton =1000000000000.0;
    Double metrictons =.907186;
    
    //initiate variables
    List <String[]> records = new ArrayList<String[]>();
    String [] nextLine = {};
    String [] nextLine2 = {};
    String [] nextLine3 = {};
    String [] nextLine4 = {};
    Double progress = 0.0;
    String [] line = {};
    List <String[]> tcpop = new ArrayList<String[]>();
    List <String[]> population = new ArrayList<String[]>();
    
    //output file name
    StringWriter sw = new StringWriter();
    CsvWriter writer = new CsvWriter("EIA-EV_Gens.csv");
    writer.setDelimiter(',');
    
    //skip first lines (titles)
    reader.readRecord();
    reader2.readRecord();
    reader3.readRecord();
    reader4.readRecord();
    
    //write titles (first lines)
    String [] title1 = reader.getValues();
    String [] title2 = {"damages"};
    writer.writeRecord(concat(title1,title2));
    
    //store pop in array
    while (reader3.readRecord()){
      nextLine3 = reader3.getValues(); 
      population.add(nextLine3);
    }
    
    //store TC in array
    Double progress2 = 0.0;
    int track = 0;
    Double sumnox = 0.0;
    Double sumso2 = 0.0;
    Double currentpop = 0.0;
    records = new ArrayList<String[]>();
    while (reader2.readRecord()){
      progress2++; track++;
      nextLine2 = reader2.getValues();
      int a = Integer.parseInt(nextLine2[2]);
      currentpop = Double.parseDouble(population.get(a-1)[1]);
      
      try {
        sumnox  = sumnox + currentpop * Double.parseDouble(nextLine2[5]);
        sumso2  = sumso2 + currentpop * Double.parseDouble(nextLine2[4]);
      }catch (ArrayIndexOutOfBoundsException e){};
      
      if(track ==3080){
        String[] d = {sumso2+"",sumnox+""};
        records.add(d);
        sumnox = 0.0;
        sumso2 = 0.0;
        track = 0;
      }
      
      
      if(progress2%100000==0){
        System.out.println("TC are " +progress2/(4*3080*3080)*100+"% copied");
      }
    }
    String[] d = {sumso2+"",sumnox+""};
    records.add(d);
    
    //this loops through the EIA-EPA data
    while (reader.readRecord()){
      //keep track of progress
      progress++;
      if(true){//progress%100==0){
        Double percent = progress/126.2;
        percent = (double)Math.round(percent * 10000) / 13796;
        System.out.println(percent+"% done");
      }
      
      
      // nextLine[] is an array of values from the current line
      //records = new ArrayList<String[]>();
      nextLine = reader.getValues();
      
      //useful values from current unit
      Double fips;
      try {fips = Double.parseDouble(nextLine[135]);}catch (NumberFormatException e){fips=0.0;};
      Double mwh1;
      try {mwh1 = Double.parseDouble(nextLine[100]);}catch (NumberFormatException e){mwh1=0.0;};
      Double mwh2; 
      try {mwh2 = Double.parseDouble(nextLine[101])/6.77;}catch (NumberFormatException e){mwh2=0.0;};
      Double capacity;
      try {if(Double.parseDouble(nextLine[11])>0.0){capacity = Double.parseDouble(nextLine[11]);}
        else{capacity = Double.parseDouble(nextLine[11]);};}catch (NumberFormatException e){
          try {capacity=Double.parseDouble(nextLine[9]);}catch (NumberFormatException f){
            capacity=Double.parseDouble(nextLine[319]);};};
        Double esh;
        try {esh = Double.parseDouble(nextLine[35]);}catch (NumberFormatException e){
          try{esh = Double.parseDouble(nextLine[36]);}catch(NumberFormatException f){esh=0.0;};};
        Double age;
        try {age = 2012.0-Double.parseDouble(nextLine[16]);}catch (NumberFormatException e){age=40.94;};
        
        //calculate MWH
        Double mwh = 0.0;
        if(mwh1>0){
          mwh=mwh1;
        }else if(mwh2>0){
          mwh=mwh2;
        }else {
          //estimate capacity if missing
          mwh = capacity *2531;
        }
        
        //calculate stack height category
        int eshcategory = 0;
        nextLine[36]=esh+"";
        if (esh==0.0){
          //based on regression of esh vs capacity
          esh=106.59+capacity*0.586+age*3.149;
          nextLine[36]=esh+"";
        }
        if(esh>500){
          eshcategory = 4;
        }else if(esh>250){
          eshcategory = 3;
        }else {
          eshcategory = 2;
        }
        
        //calculate heat rate
        Double hr =10.4;
        hr = Double.parseDouble(nextLine[483]);
        
        // read through TC file
        reader2 = new CsvReader(new FileReader(ADDRESS_FILE2));
        
        //go through pop file and find index (1-3080) of current fips code
        reader3 = new CsvReader(new FileReader(ADDRESS_FILE3));
        reader3.readRecord();
        Double index = 0.0;
        while (reader3.readRecord()){
          nextLine3 = reader3.getValues(); 
          if(Double.parseDouble(nextLine3[0])==fips){
            index = Double.parseDouble(nextLine3[2]);
          }
        }
        
        Double sum = 0.0;
        
        // add up s02
        Double so2 = 0.0;
        Double nox = 0.0;
        
        if(!(index==0.0)){
          so2 = Double.parseDouble(records.get((int)((index-1)*4+(eshcategory)-1))[0]);
          if(so2==0.0){
            so2 = Double.parseDouble(records.get((int)((index-1)*4+(1)-1))[0]);
          }
          nox = Double.parseDouble(records.get((int)((index-1)*4+(eshcategory-1)))[1]);
          if(nox==0.0){
            nox = Double.parseDouble(records.get((int)((index-1)*4+(1-1)))[1]);
          }
        }
        
        //emissions
        Double emissionso2 =0.0;
        emissionso2 = Double.parseDouble(nextLine[480])*hr*mwh/2000;
        
        //biomass unit with missing emissions
        try{
          if(nextLine[148].equals("Wood or Wood Waste")){
            emissionso2=0.0000519*hr*mwh;
          } }catch (ArrayIndexOutOfBoundsException f){}
          
          if(emissionso2==0.0){try{
            if(nextLine[88].equals("COAL")){
              //coal unit with missing emissions
              emissionso2=0.000459132*mwh*1.1023;
              //gas unit with missing emissions
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CC")){
              emissionso2=0.000000392165*mwh*1.1023;
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CE")){
              emissionso2=0.0000003*mwh*1.1023;
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CS")){
              emissionso2=0.000000303*mwh*1.1023;
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CT")){
              emissionso2=0.000000585321*mwh*1.1023;
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("GT")){
              emissionso2=0.000007938984*mwh*1.1023;
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("IC")){
              emissionso2=0.000007938984*mwh*1.1023;
            }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("ST")){
              emissionso2=0.00000919132*mwh*1.1023;
              //liquid unit with missing emissions
            }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("CC")){
              emissionso2=0.000004264971*mwh*1.1023;
            }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("CT")){
              emissionso2=0.000006365628*mwh*1.1023;
            }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("GT")){
              emissionso2=0.000127312561*mwh*1.1023;
            }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("IC")){
              emissionso2=0.0008*mwh*1.1023;
            }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("ST")){
              emissionso2=0.000231160079*mwh*1.1023;
            }else{
              emissionso2=0.0;
            }
            nextLine[480]=emissionso2*2000.0/(mwh*hr)+"";
          }catch(NumberFormatException e){}
          }
          
          Double emissionnox = 0.0;
          emissionnox = Double.parseDouble(nextLine[479])*hr*mwh/2000;
          
          try{        
            //biomass unit with missing emissions
            if(nextLine[148].equals("Wood or Wood Waste")){
              emissionnox=0.0000884507*mwh*1.1023;
            } }catch (ArrayIndexOutOfBoundsException f){}

            if(emissionnox == 0.0){
              try{
                if(nextLine[88].equals("COAL")){
                  //coal unit with missing emissions
                  emissionnox=0.000132*mwh*1.1023;
                  //gas unit with missing emissions
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CC")){
                  emissionnox=0.000013700338*mwh*1.1023;
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CE")){
                  emissionnox=0.000018032892*mwh*1.1023;
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CS")){
                  emissionnox=0.000000303298*mwh*1.1023;
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("CT")){
                  emissionnox=0.000000585321*mwh*1.1023;
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("GT")){
                  emissionnox=0.000007938984*mwh*1.1023;
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("IC")){
                  emissionnox=0.0055*mwh*1.1023;
                }else if(nextLine[88].equals("GAS")&&nextLine[9].equals("ST")){
                  emissionnox=0.00000919132*mwh*1.1023;
                  //liquid unit with missing emissions
                }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("CC")){
                  emissionnox=0.000039365409*mwh*1.1023;
                }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("CT")){
                  emissionnox=0.000058754342*mwh*1.1023;
                }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("GT")){
                  emissionnox=0.000318329*mwh*1.1023;
                }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("IC")){
                  emissionnox=0.0135*mwh*1.1023;
                }else if(nextLine[88].equals("LIQUID")&&nextLine[9].equals("ST")){
                  emissionnox=0.000105024*mwh*1.1023;
                }else{
                  emissionnox=0.0;
                }
                nextLine[479]=emissionnox*2000.0/(mwh*hr)+"";
                
              }
              catch(NumberFormatException e){};}
            
            sum = (so2*so2coef*emissionso2+nox*noxcoef*emissionnox)*.593;
            
            // calculate damages
            Double damages = 0.0;
            damages = beta * y * life * (mgperton / secperyear) * metrictons * sum;
            Double dollarpermwh = 0.0;
            dollarpermwh = damages / mwh;
            
            //output columns
            int cols = 487;
            nextLine[cols] = dollarpermwh*mwh + "";
            nextLine[cols-1] = dollarpermwh + "";
            nextLine[cols-2] = emissionso2 + "";
            nextLine[cols-3] = emissionnox + "";
            nextLine[cols-5] = mwh + "";
            nextLine[cols-6] = sum + "";
            
            //write results
            writer.writeRecord(nextLine);
            writer.flush();
    }
    
    writer.flush();
    writer.close();
    
    //print results of match
    System.out.println("Done" );
    
  }
  //function to concatenate string arrays
  public static final String[] concat(String[] s1, String[] s2) {
    String[] erg = new String[s1.length + s2.length];
    System.arraycopy(s1, 0, erg, 0, s1.length);
    System.arraycopy(s2, 0, erg, s1.length, s2.length);
    return erg;
  }
}

