//**********************************************************
// Program: match.java
//
// Author: Charles Marquet, cmm363@cornell.edu
//
// Description: This program reads the EIA and EPA data files and tries to match 
//              the generator data based on generator ID, capacity and type.
//
//              The program goes through each EIA unit one at a time. For each EIA unit:
//               1) It finds the set of units in the EPA data corresponding to the same 
//                  plant code as the current EIA unit.
//               2) It assigns a score to each of the EPA units in the set, based on the similarities
//                  the unit matches the current EIA unit
//               3) The EPA unit with the highest score is matched with the current EIA unit,
//                  as long as it passes some validity tests (e.g. close enough capacities)
//              
// Input: The EIA and EPA data sets, csv formats      
//        The EIA data is first matched with stacks data
//        These files require added columns from the raw data, such as fuel type and EPA est. capacity
//
// Output: The match.csv file, with all EIA units and the matched EPA units.
//
// Need the free csv reader/writer from: http://sourceforge.net/projects/javacsv/
//*********************************************************

import java.io.*;
import java.lang.String;
import java.lang.Math;
import java.lang.Integer;
import com.csvreader.CsvReader;
import com.csvreader.CsvWriter;
import java.util.regex.*;
import java.util.*;


public class match {
  
  //this is the EIA and EPA files, we are using the EIA as the master spreadsheet.
  private static final String ADDRESS_FILE="EIA.csv";
  private static final String ADDRESS_FILE2="EPA.csv";
  private static final String ADDRESS_FILE3="county_fips.csv";
  private static final String ADDRESS_FILE4="plantdata.csv";
  
  public static void main(String[] args) throws IOException {
    
    //these read the input files
    CsvReader reader = new CsvReader(new FileReader(ADDRESS_FILE));
    CsvReader reader2= new CsvReader(new FileReader(ADDRESS_FILE2));
    CsvReader reader3= new CsvReader(new FileReader(ADDRESS_FILE3));
    CsvReader reader4= new CsvReader(new FileReader(ADDRESS_FILE4));
    
    //initiate variables
    List <String[]> records = new ArrayList<String[]>();
    String [] nextLine = {};
    String [] nextLine2 = {};
    String [] nextLine3 = {};
    String [] nextLine4 = {};
    String currentPlantEIA = "";
    String currentGenEIA = "";
    Double currentGenEIA_capacity =0.0;
    String currentGenEIA_type = "";
    String currentGenEIA_generatorID = "";
    String [] line ={};
    
    //variables to keep track of progress and output
    int nomatch=0; Double progress=0.0;Double fossilcapacity =0.0; 
    int count=0; int count1=0;int count2=0;int count3=0; int counteia = 0;
    Double epacapacity =0.0;Double epacapacity1 =0.0;Double epacapacity2 =0.0;Double epacapacity3 =0.0;
    
    StringWriter sw = new StringWriter();
    //output file name
    CsvWriter writer = new CsvWriter("match_fips.csv");
    CsvWriter writer1 = new CsvWriter("EIA_mod_full.csv");
    writer.setDelimiter(',');
    writer1.setDelimiter(',');
    
    //skip first lines (titles)
    reader.readRecord();
    reader2.readRecord();
    
    //write titles to output file
    String [] title1 = reader.getValues();
    String [] title2 = reader2.getValues();
    String [] add_fuel = {"Fuel_Type","Latitude","Longitude"};
    
    writer.writeRecord(concat(concat(title1,add_fuel),title2));
    writer1.writeRecord(title1);
    
    //add fuel type to EIA 
    while (reader.readRecord()){
      String [] add1 = {"","",""};
      nextLine = concat(reader.getValues(),add1);

      //add fuel type
      if(nextLine[15].equals("BIT")||nextLine[15].equals("ANT")||nextLine[15].equals("LIG")||
         nextLine[15].equals("RC")||nextLine[15].equals("SUB")||nextLine[15].equals("WC")){
        nextLine[86]="COAL";
      }else if(nextLine[15].equals("DFO")||nextLine[15].equals("RFO")||nextLine[15].equals("JF")||
               nextLine[15].equals("KER")||nextLine[15].equals("BLQ")||nextLine[15].equals("OBL")||nextLine[15].equals("WDL")||nextLine[15].equals("WO")){
        nextLine[86]="LIQUID";
      }else if(nextLine[15].equals("PC")){
        nextLine[86]="Petroleum Coke";
      }else if(nextLine[15].equals("TDF")){
        nextLine[86]="TIRE";
      }else if(nextLine[15].equals("AB")||nextLine[15].equals("MSW")||nextLine[15].equals("OBS")||
               nextLine[15].equals("SLW")||nextLine[15].equals("WDS")){
        nextLine[86]="OTHER SOLID";
      }else if(nextLine[15].equals("NG")||nextLine[15].equals("BFG")||nextLine[15].equals("LFG")||
               nextLine[15].equals("OBG")||nextLine[15].equals("OG")||nextLine[15].equals("PG")||
               nextLine[15].equals("SG")||nextLine[15].equals("SGC")||nextLine[28].equals("Y")){
        nextLine[86]="GAS";
      }else {
        nextLine[86]="N/A";
      }
      
      writer1.writeRecord(nextLine);
    }
    writer1.flush();
    writer1.close();
    
    reader = new CsvReader(new FileReader("EIA_mod_full.csv"));
    reader.readRecord();
    
    //this loops through the EIA data, one unit at a time
    while (reader.readRecord()){
      //keep track of progress, and print % done as program runs
      progress++;
      if(progress%100==0){
        Double percent = progress/164;
        percent = (double)Math.round(percent * 10000) / 10000;
        System.out.println(percent+"% done");
      }
      
      // nextLine[] is the array of values from the current line (this is all of the data from the current EIA unit)
      records = new ArrayList<String[]>();
      nextLine = reader.getValues();
      
      //store EIA data from current line into variables
      currentPlantEIA = nextLine[2];
      currentGenEIA = nextLine[6];
      currentGenEIA_type = nextLine[86];
      
      //The variable holding the value for the EIA units' capacity
      // replace the first 9 by 11 in order to use winter capacity instead of nameplace capacity as a default
      try {  currentGenEIA_capacity = Double.parseDouble(nextLine[11]);}
      catch (NumberFormatException e){currentGenEIA_capacity=Double.parseDouble(nextLine[9]);};
      
      //Account for EIA winter capacity possible error
      if(currentGenEIA_capacity < 0.0001){
        currentGenEIA_capacity=Double.parseDouble(nextLine[9]); 
      }
      
      //STEP 1: finds the set of units in the EPA data corresponding to the same plant code as the current EIA unit.
      reader2 = new CsvReader(new FileReader(ADDRESS_FILE2));
      
      //loop through the all the EPA data 
      //only if EIA plant is using fuel, transmit to the grid, is not out of service and is in the mainland 48 states
      if((currentGenEIA_type.equals("COAL")||currentGenEIA_type.equals("GAS")||currentGenEIA_type.equals("LIQUID")||
          currentGenEIA_type.equals("TIRE")||currentGenEIA_type.equals("OTHER SOLID")||currentGenEIA_type.equals("Petroleum Coke"))&&
         currentGenEIA_capacity>0&&!nextLine[8].equals("OS")&&!nextLine[7].equals("CA")&&
         !nextLine[4].equals("AK")&&!nextLine[4].equals("HI")){
        fossilcapacity = fossilcapacity + currentGenEIA_capacity;
        counteia=counteia+1;
        
        while (reader2.readRecord()){
          nextLine2 = reader2.getValues();
          String a = nextLine2[2];
          //store the matching units in arraylist (units with same plant code as current EIA unit)
          if(a.equals(currentPlantEIA)){
            Double z;
            try {z = Double.parseDouble(nextLine2[9]);} catch (NumberFormatException e){
              try {z = Double.parseDouble(nextLine2[10])/6.77;} catch (NumberFormatException f){z=0.0;};};
            Double y;
            try {y = Double.parseDouble(nextLine2[13]);} catch (NumberFormatException e){y=-1.0;};
            if((z>=0)){// add this to the if statement to the left if not same year of EPA and EIA datatsets: &&nextLine2[27].equals("Operating")
              //narrow matchable sets for gas
              if (currentGenEIA_type.equals("GAS")){
                if (nextLine[7].equals("CS")||nextLine[7].equals("CT")){
                  //combustion turbine part of combined cycle set
                  records.add(nextLine2);
                }
                else if (nextLine[7].equals("CA")){
                  //steam part of combined cycle set
                }
                else if (nextLine[7].equals("GT")){
                  //combustion turbine not part of combined cycle set
                  records.add(nextLine2);
                }
                else {
                  records.add(nextLine2);
                }
              }//narrow matchable set for liquids
              else if (currentGenEIA_type.equals("LIQUID")){
                if (nextLine[7].equals("CS")||nextLine[7].equals("CT")){
                  //combustion turbine part of combined cycle set
                  records.add(nextLine2);
                }
                else if (nextLine[7].equals("CA")){
                  //steam part of combined cycle set
                }
                else if (nextLine[7].equals("GT")){
                  //combustion turbine not part of combined cycle set
                  records.add(nextLine2);
                }
                else {
                  //other
                  records.add(nextLine2);
                }
              }
              else {
                records.add(nextLine2);
              }
            }
          }
        }
      }
      
      //PART 2: score matching units
      int size = records.size();
      Double[] score = null;
      if (size>0){
        score = new Double[size];
        Pattern intsOnly = Pattern.compile("\\d+");
        //makematch gets the integers only out of the generator_id (EIA)
        Matcher makeMatch = intsOnly.matcher(nextLine[6]);
        String number = "";
        while (makeMatch.find()) {
          number=number+makeMatch.group();
        }
        
        Pattern lettersOnly = Pattern.compile("\\D+");
        //makematch gets the letters only out of the generator_id (EIA)
        Matcher makeMatch3 = lettersOnly.matcher(nextLine[6]);
        String letter = "";
        while (makeMatch3.find()) {
          letter=letter+makeMatch3.group();
        }
        
        int intEIA ;
        try{intEIA= Integer.parseInt(number);}catch(NumberFormatException e){intEIA=-1;};
        currentGenEIA_generatorID=nextLine[6];
        
        for (int i=1; i<size+1; i++){
          //makematch gets the ints only out of the generator_id (EPA)
          Matcher makeMatch2 = intsOnly.matcher(records.get(i-1)[3]);
          String number2 = "";
          while (makeMatch2.find()) {
            number2=number2+makeMatch2.group();
          }
          
          //makematch gets the letters only out of the generator_id (EPA)
          Matcher makeMatch4 = lettersOnly.matcher(records.get(i-1)[3]);
          String letter2 = "";
          while (makeMatch4.find()) {
            letter2=letter2+makeMatch4.group();
          }
          
          int intEPA ;
          try{intEPA= Integer.parseInt(number2);}catch (NumberFormatException e){intEPA=-2;};
          
          //Assign score to EPA unit, based on information available
          //The numbers used are somewhat arbitrary, but are intended to give matching units the highest score
          
          //there is negative points based on of different the capacities are
          score[i-1]=-500*Math.max((Math.abs(currentGenEIA_capacity-Double.parseDouble(records.get(i-1)[38]))/currentGenEIA_capacity),
                                   Math.abs(currentGenEIA_capacity-Double.parseDouble(records.get(i-1)[38]))/Double.parseDouble(records.get(i-1)[38]));
          //a large positive score if generator ids match exactly
          if(currentGenEIA_generatorID.equals(records.get(i-1)[3])){score[i-1]=score[i-1]+1000;}
          // a smaller positive score if the generator ids match by number (e.g. "CT1" and "1")
          if(intEIA==intEPA){score[i-1]=score[i-1]+500;}
          // a smaller positive score if the generator ids match by letters (e.g. "1A" and "A")
          if(letter.equals(letter2)){score[i-1]=score[i-1]+500;}
          // a negative score if the generator types are different
          if(!((records.get(i-1)[37]).equals(currentGenEIA_type))){score[i-1]=score[i-1]-350;}
        }
      }
      
      //PART 3:select record with max score and that meets conditions & write it to output file
      // conditions are: 
      // 1)capacities are within half and double of each other (e.g. if EIA is 100MW, EPA needs to be within 50MW and 200MW)
      // 2)generation types are the same (e.g. "COAL" and "COAL")
      // based on trial and error, there has been additional rules added
      Double max = -10000000.0; int maxi = 0; Boolean empty = true;
      if(size>0){
        for (int i = 0; i < score.length; i++){
          if(max < score[i]){
            // for units with <25MW score must be >-350
            if(currentGenEIA_capacity<=25){
              if(score[i]>-350&&records.get(i)[37].equals(currentGenEIA_type)&&
                 (((currentGenEIA_capacity<Double.parseDouble(records.get(i)[38])*2) &&
                   (currentGenEIA_capacity>Double.parseDouble(records.get(i)[38])/2))||Double.parseDouble(records.get(i)[38])==0)){
                empty = false;
                max = score[i];
                maxi = i; 
              }
            }
            //normal units
            else if(records.get(i)[37].equals(currentGenEIA_type)&&
                    (((currentGenEIA_capacity<Double.parseDouble(records.get(i)[38])*4) &&
                      (currentGenEIA_capacity>Double.parseDouble(records.get(i)[38])/4))||Double.parseDouble(records.get(i)[38])==0)){
              empty = false;
              max = score[i];
              maxi = i;
            }
            //units with data errors in generation types
            else if(line[37].equals("GAS")&&currentGenEIA_type.equals("LIQUID")&&line[30].equals("Diesel Oil")&&
                    ((currentGenEIA_capacity<Double.parseDouble(line[38])*4) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/4))){
              empty = false;
              max = score[i];
              maxi = i;
            }
            else if(line[37].equals("LIQUID")&&currentGenEIA_type.equals("GAS")&&line[30].equals("Pipeline Natural Gas")&&
                    ((currentGenEIA_capacity<Double.parseDouble(line[38])*4) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/4))){
              empty = false;
              max = score[i];
              maxi = i;
            }
          }
        }
        
        // if there is no max score found from above try again with no rules
        if(empty){
          for (int i = 0; i < score.length; i++){
            if(max < score[i]){
              empty = false;
              max = score[i];
              maxi = i;
            }
          }
        }
        // store unit with highest score into variable
        line = records.get(maxi);
      }
      
      //flag variable (to keep track of units matched and units unmatched)
      String[] flag = new String[2]; 
      String[] blank = new String[45];
      flag[0]="1";
      String[] maxscore = new String[1]; 
      maxscore[0]=max+"";
      Boolean fipsfound = false;
      
      if(line.length>0){
        //find fips code from file
        reader3= new CsvReader(new FileReader(ADDRESS_FILE3));
        while (reader3.readRecord()){
          nextLine3 = reader3.getValues();
          if(nextLine3[1].equals(line[0])&&nextLine3[0].equals(line[18])){
            flag[1]=nextLine3[2];
            fipsfound = true;
          }
        }
      }else{
        reader3= new CsvReader(new FileReader(ADDRESS_FILE3));
        while (reader3.readRecord()){
          nextLine3 = reader3.getValues();
          if(nextLine3[1].equals(nextLine[4])&&nextLine3[0].equals(nextLine[5])){
            flag[1]=nextLine3[2];
            fipsfound = true;
          }
        }
      }
      if(fipsfound==false){
        reader3= new CsvReader(new FileReader(ADDRESS_FILE3));
        while (reader3.readRecord()){
          nextLine3 = reader3.getValues();
          if(nextLine3[1].toUpperCase().equals(nextLine[4].toUpperCase())&&nextLine3[0].toUpperCase().equals(nextLine[5].toUpperCase())){
            flag[1]=nextLine3[2];
            fipsfound = true;
          }
        }
      }
      if(fipsfound==false){
        reader3= new CsvReader(new FileReader(ADDRESS_FILE3));
        while (reader3.readRecord()){
          nextLine3 = reader3.getValues();
          if(nextLine3[1].equals(nextLine[4])){
            flag[1]=nextLine3[2];
            fipsfound = true;
          }
        }
      }
      //find latitude and longitude (set to average value if missing)
      reader4= new CsvReader(new FileReader(ADDRESS_FILE4));
        while (reader4.readRecord()){
          nextLine4 = reader4.getValues();
          if(nextLine4[1].equals(nextLine[2])){
            try{Double e = Double.parseDouble(nextLine4[8]);
              if(e>0){
                nextLine[87]=nextLine4[8];
              }else{
                nextLine[87]="38";
              }
              
            }catch(NumberFormatException e){
            nextLine[87]="38";
            }
            try{Double e = Double.parseDouble(nextLine4[9]);
              if(e<0){
            nextLine[88]=nextLine4[9];
              }else{
                nextLine[88]="-97";
              }
            }catch(NumberFormatException e){
            nextLine[88]="-97";
            }
            nextLine[32]=nextLine4[11];
          }
        }
        
      // write line to csv, with flag variable to indicate type of match
      // only if it meets minimum criteria:
      // 1)capacities are within half and double of each other 
      // 2)generation types are the same (e.g. "COAL" and "COAL")
      // based on trial and error, there has been additional rules added      
      
      if(size>0&&line[37].equals(currentGenEIA_type)&&
         (((currentGenEIA_capacity<Double.parseDouble(line[38])*2) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/2))||Double.parseDouble(line[38])==0)){
        //Meets normal criteria, EPA generator matched to EIA generator
        count++; 
        epacapacity = epacapacity + currentGenEIA_capacity;
        flag[0]="Matched";
        line = concat(concat(line,maxscore),flag);
        writer.writeRecord(concat(nextLine,line));
      }
      else if(size>0&&line[37].equals("GAS")&&currentGenEIA_type.equals("LIQUID")&&line[30].equals("Diesel Oil")&&
              ((currentGenEIA_capacity<Double.parseDouble(line[38])*4) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/4))){
        //Doesn't meet criteria (liquid/gas mistmatched, but by data error), EPA generator matched to EIA generator
        count++; 
        epacapacity = epacapacity + currentGenEIA_capacity;
        flag[0]="Matched";
        line = concat(concat(line,maxscore),flag);
        writer.writeRecord(concat(nextLine,line));
      }
      else if(size>0&&line[37].equals("LIQUID")&&currentGenEIA_type.equals("GAS")&&line[30].equals("Pipeline Natural Gas")&&
              ((currentGenEIA_capacity<Double.parseDouble(line[38])*4) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/4))){
        //Doesn't meet criteria (liquid/gas mistmatched, but by error), EPA generator matched to EIA generator
        count++; 
        epacapacity = epacapacity + currentGenEIA_capacity;
        flag[0]="Matched";
        line = concat(concat(line,maxscore),flag);
        writer.writeRecord(concat(nextLine,line));
      }
      else if(size>0&&currentGenEIA_capacity<=25.0&&line[37].equals(currentGenEIA_type)&&max>-30&&
              ((currentGenEIA_capacity<Double.parseDouble(line[38])*4) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/4))){
        //Doesn't meet criteria (but EPA capacity 1/4 to 4 times EIA, type matches, and score >0), EPA generator matched to EIA generator
        count++; 
        epacapacity = epacapacity + currentGenEIA_capacity;
        flag[0]="Matched";
        line = concat(concat(line,maxscore),flag);
        writer.writeRecord(concat(nextLine,line));
      }
      else if(size>0&&currentGenEIA_capacity>25.0&&line[37].equals(currentGenEIA_type)&&
              ((currentGenEIA_capacity<Double.parseDouble(line[38])*4) &&(currentGenEIA_capacity>Double.parseDouble(line[38])/4))){
        //Doesn't meet criteria (but EPA capacity 1/4 to 4 times EIA, type matches)
        count1++; 
        epacapacity = epacapacity + currentGenEIA_capacity;
        flag[0]="Matched";
        line = concat(concat(line,maxscore),flag);
        writer.writeRecord(concat(nextLine,line));
      }
      else if(size>0&&line[37].equals(currentGenEIA_type)){
        //Doesn't meet criteria (EPA capacity not within 1/4 to 4 times EIA, type matches), EPA generator matched to EIA generator
        count2++; 
        epacapacity2 = epacapacity2 + currentGenEIA_capacity;
        flag[0]="Not matched";
        line = concat(concat(line,maxscore),flag);
        blank[44]=flag[1];
        writer.writeRecord(concat(nextLine,blank));
      }
      else if(size>0&&!line[37].equals(currentGenEIA_type)){
        //Doesn't meet criteria (type does not match), EPA generator matched to EIA generator
        count3++; 
        epacapacity3 = epacapacity3 + currentGenEIA_capacity;
        flag[0]="Not matched";
        line = concat(concat(line,maxscore),flag);
        blank[44]=flag[1];
        writer.writeRecord(concat(nextLine,blank));
      }
      else{
        //no match
        nomatch++;
        blank[44]=flag[1];
        writer.writeRecord(concat(nextLine,blank));
      }
      //write units to file
      writer.flush();
      
    }
    //write units to file & close file
    writer.flush();
    writer.close();
    
    //print summary result of match
    System.out.println("Done" );
    System.out.println("Units matched that met criteria: "+count );
    System.out.println("Total EIA fuel capacity: "+fossilcapacity+", Total EPA matched fuel capacity: "+epacapacity);
  }
  
  //function to concatenate string arrays
  public static final String[] concat(String[] s1, String[] s2) {
    String[] erg = new String[s1.length + s2.length];
    System.arraycopy(s1, 0, erg, 0, s1.length);
    System.arraycopy(s2, 0, erg, s1.length, s2.length);
    return erg;
  }
}

