//**********************************************************
// Program: regression_prepare.java
//
// Author: Charles Marquet
//
// Description: This program prepares data for regression software input
//              
// Input: The EIA-EV_Gens_1.csv file
//
// Output: The regressionDatafull.csv file, ready for regression analysis in Stata
//
// Need the free csv reader/writer from: http://sourceforge.net/projects/javacsv/
//*********************************************************

import java.io.*;
import java.lang.String;
import java.lang.Math;
import java.lang.Integer;
import com.csvreader.CsvReader;
import com.csvreader.CsvWriter;
import java.util.regex.*;
import java.util.*;


public class regression_prepare_full {
  //these are the input file names
  //Input: EIA-EPA-stacks matched file
  private static final String ADDRESS_FILE="EIA-EV_Gens_1.csv";
  
  public static void main(String[] args) throws IOException {
    
    //these read the input files
    CsvReader reader = new CsvReader(new FileReader(ADDRESS_FILE));
    
    //constants
    int year = 2010;
    
    //initiate variables
    String [] nextLine = {};
    String [] line = new String[57];
    
    //output file name
    StringWriter sw = new StringWriter();
    CsvWriter writer = new CsvWriter("regressionDatafull.csv");
    writer.setDelimiter(',');
    
    //write title
    String [] title = {"ID","prime mover","energy source","d_dpt","nameplate","age","mult_fuels","sfg_system",
      "nox_rate","so2_rate","heat_rate","matched_to_epa","lat","longitude","fgd","caa1","caa2","caa3","low1",
      "low2","low3","c_CF","c_CU","c_IF","c_NC","c_ND","c_OT","c_SS","c_SU","c_TU","c_US","c_WA",
      "l_AA","l_BF","l_CF","l_FR","l_FU","l_H2","l_LA","l_LN","l_NA","l_NC","l_NH","l_OT","l_OV","l_SN",
      "l_SR","l_ST","PE","pec1","pec2","pec3","pel1","pel2","pel3","gload","sload"};
    writer.writeRecord(title);
    
    //skip first line
    reader.readRecord();
    int c =1;
    
    //this loops through the EIA-EPA data
    while (reader.readRecord()){
      nextLine = reader.getValues(); 
      if(nextLine[2].equals("")){
        line[0]=c+""; 
        c++;
        line[11]="0";
        writer.writeRecord(line);
        writer.flush();
        
      }else{
        line[0]=c+""; 
        c++;
        line[1]=nextLine[9];
        line[2]=nextLine[17];
        line[3]="0";
        if(nextLine[20].equals("Y")){
          line[3]="1";
        }
        
        //numerical variables
        line[4]=nextLine[11];
        line[5]=Math.max((year-Integer.parseInt(nextLine[16])),1)+"";
        //binary variables
        if(nextLine[19].equals("Y")){line[6]="1";}else{line[6]="0";};
        if(nextLine[30].equals("Y")){line[7]="1";}else{line[7]="0";};
        
        //explanatory variables
        line[8]=nextLine[103];
        try{line[9]=(Double.parseDouble(nextLine[102])/Double.parseDouble(nextLine[106])*2000)+"";}catch(NumberFormatException e){line[9]="";};
        line[10]=nextLine[127];
        //is the unit matched to epa?
        if(nextLine[91].equals("")){line[11]="0";}else{line[11]="1";};
        
        //lat and long
        line[12]=nextLine[89];
        line[13]=nextLine[90];
        
        //fgd status
        line[14]="1";
        if(nextLine[56].equals("")){
          line[14]="0";
        }
        
        String caa1;
        String caa2;
        String caa3;
        String low1;
        String low2;
        String low3;
        
        //CAA
        try{caa1=nextLine[50].substring(0,2);}catch(StringIndexOutOfBoundsException e){caa1="N/A";};
        if(caa1.equals("||")){
          caa1=nextLine[50].substring(2,4);
        }
        try{caa2=nextLine[51].substring(0,2);}catch(StringIndexOutOfBoundsException e){caa2="N/A";};
        if(caa2.equals("||")){
          caa2=nextLine[51].substring(2,4);
        }
        try{caa3=nextLine[52].substring(0,2);}catch(StringIndexOutOfBoundsException e){caa3="N/A";};
        if(caa3.equals("||")){
          caa3=nextLine[52].substring(2,4);
        }
        
        //LOW NOX
        try{low1=nextLine[85].substring(0,2);}catch(StringIndexOutOfBoundsException e){low1="N/A";};
        if(low1.equals("||")){
          low1=nextLine[85].substring(2,4);
        }
        try{low2=nextLine[86].substring(0,2);}catch(StringIndexOutOfBoundsException e){low2="N/A";};
        if(low2.equals("||")){
          low2=nextLine[86].substring(2,4);
        }
        try{low3=nextLine[87].substring(0,2);}catch(StringIndexOutOfBoundsException e){low3="N/A";};
        if(low3.equals("||")){
          low3=nextLine[87].substring(2,4);
        }
        line[15]=caa1;
        line[16]=caa2;
        line[17]=caa3;
        line[18]=low1;
        line[19]=low2;
        line[20]=low3;
        
        String [] caanames = {"CF","CU","IF","NC","ND","OT","SS","SU","TU","US","WA"};
        int t = 0;
        for(int d=0;d<11;d++){
          line[t+21]="0";
          if(caa1.equals(caanames[d])||caa2.equals(caanames[d])||caa3.equals(caanames[d])){
            line[t+21]="1";
          }
          t++;
        }
        String [] lownames = {"AA","BF","CF","FR","FU","H2","LA","LN","NA","NC","NH","OT","OV","SN","SR","ST"};
        t = 0;
        for(int d=0;d<16;d++){
          line[t+32]="0";
          if(low1.equals(lownames[d])||low2.equals(lownames[d])||low3.equals(lownames[d])){
            line[t+32]="1";
          }
          t++;
        }
        
        
        //PE columns
        line[48]=nextLine[9]+"_"+nextLine[17];
        
        //PEC columns
        line[49]=nextLine[9]+"_"+nextLine[17]+"_"+line[15];
        line[50]=nextLine[9]+"_"+nextLine[17]+"_"+line[16];
        line[51]=nextLine[9]+"_"+nextLine[17]+"_"+line[17];
        
        //PEL columns
        line[52]=nextLine[9]+"_"+nextLine[17]+"_"+line[18];
        line[53]=nextLine[9]+"_"+nextLine[17]+"_"+line[19];
        line[54]=nextLine[9]+"_"+nextLine[17]+"_"+line[20];
        
        //GLOAD+SLOAD
        line[55]=nextLine[100];
        line[56]=nextLine[101];
        
        writer.writeRecord(line);
        writer.flush();
      }
    }
    
    writer.flush();
    writer.close();
    
    //print results of match
    System.out.println("Done" );
    
  }
  //function to concatenate string arrays
  public static final String[] concat(String[] s1, String[] s2) {
    String[] erg = new String[s1.length + s2.length];
    System.arraycopy(s1, 0, erg, 0, s1.length);
    System.arraycopy(s2, 0, erg, s1.length, s2.length);
    return erg;
  }
}

